@@ -22,7 +22,7 @@ lg.setLevel(4)
 
 class PolyMaker:
     def __init__(self):
-        self.smiles_req = {
+        self.smiles_templates = {
             "ols": "[C,c;!$(C=O)][OH]",
             "aliphatic_ols": "[C;!$(C=O);!$([a])][OH]",
             "acids": "[#6][#6](=[#8:4])([F,Cl,Br,I,#8H,O-])",
@@ -31,6 +31,14 @@ class PolyMaker:
             "cyclic_carbonates": "[O]=[C]1[O][C][C][O]1",
             "acidanhydrides": "[#8]([#6](=[#8]))([#6](=[#8]))",
             "prime_thiols": "[#6;!$(C=O)][SH]",
+            "isocyanates": "[#6][#7;!$([#7+])]=[#6]=[#8]",
+            "acrylates": "[CH2]=[C][C](=[O])",
+            "double_bond": "C=C",
+            "vinyls": "[CH2]=[CH!$(CC=O)]",
+            "imides": "[#8:3]([#6:4](=[#8:5]))([#6:6](=[#8:7]))",
+            "double_bond_secondary": "[CH]=[CH]",
+            "double_bond_tertiary": "C[C]=[C!$([CH])]C",
+            "double_bond_quaternary": "C[C!$([CH])]=[C!$([CH])]C",
         }
         self.reactions = {
             "ester": {
@@ -84,10 +92,17 @@ class PolyMaker:
                 "[C:7][NH:6][C:1](=[O:0])[O:2][C:3][C:4][O:5]",
                 "diamine_cycliccarbonate": "[NH2;!$([NH2+]):6][#6;!$(C=O):7].[O:0]=[C:1]1[O:2][C:3][C:4][O:5]1>>"
                 "[C:7][NH:6][C:1](=[O:0])[O:2][C:3][C:4][O:5]",
-                "infinite_chaine": "to complete",
+                "infinite_chain": "to complete",
+            },
+            "urethane": {
+                "diisocyanates_ol": "[#6:0][#7;!$([#7+]):1]=[#6:2]=[#8:3].[C,c;!$(C=O):4][OH:5]>>"
+                "[#6:0][#7;!$([#7+]):1]-[#6:2](=[#8:3])[O:5][C;!$(C=O):4]",
+                "diols_isocyanate": "[C,c;!$(C=O):4][OH:5].[#6:0][#7;!$([#7+]):1]=[#6:2]=[#8:3]>>"
+                "[#6:0][#7;!$([#7+]):1]-[#6:2](=[#8:3])[O:5][C;!$(C=O):4]",
+                "infinite_chain": "to complete",
             },
         }
-        self.__verison__ = "0.1.5.0"
+        self.__verison__ = "0.1.7.0"
 
     @staticmethod
     def checksmile(smi: str) -> Union[str, None]:
@@ -323,10 +338,6 @@ class PolyMaker:
         pd.DataFrame
             A dataframe containing the polymerized inputs
         """
-        # Get rid of these two different ones. Using now just for getting it to work.
-
-        df_return_mapped = pd.DataFrame()
-
         # Ensure there is a Pm value, default is 1
         if "pm" not in reactants:
             reactants["pm"] = pm
@@ -349,18 +360,17 @@ class PolyMaker:
 
         reactants["DP"] = DP
 
-        stereo_df = reactants.copy()
-        if not stereo_df.empty:
-            stereo_df.loc[:, "monomers"] = stereo_df.monomers.astype(str)
-            df_return_mapped = pd.concat([df_return_mapped, stereo_df])
+        df_return_mapped = reactants
 
         if verbose:
             df_return_mapped = df_return_mapped.progress_apply(
-                self._polymerizemechanism_thermoplastic_stereo, axis=1,
+                self._polymerizemechanism_thermoplastic_stereo,
+                axis=1,
             )
         else:
             df_return_mapped = df_return_mapped.apply(
-                self._polymerizemechanism_thermoplastic_stereo, axis=1,
+                self._polymerizemechanism_thermoplastic_stereo,
+                axis=1,
             )
 
         if not df_return_mapped.empty:
@@ -401,7 +411,7 @@ class PolyMaker:
             # TODO: how to integrate
             return pd.DataFrame()
         else:
-            df_poly[row.keys()] = row
+            df_poly = pd.merge(df_poly, row.to_frame().T, how="cross")
             df_poly["replicate_structure"] = [
                 i for i in range(row.replicate_structures)
             ]
@@ -424,7 +434,7 @@ class PolyMaker:
         ----------
         reactants: a tuple
                    or a strings of monomers
-                   a pandas dataframe containing a column titled "monomers" (generated by get_monomers)
+                   or a pandas dataframe containing a column titled "monomers" (generated by get_monomers)
                    optional column "distribution" containing list of molar ratio that maps to monomers in "monomers" column
                    optional column mechanism containing desired mechanism to use for the row
 
@@ -432,9 +442,15 @@ class PolyMaker:
             Degree of polymerization, by default 2
 
         mechanism: string,
-            vinyl: performs polymerization along vinyl groups
-            ester: performs condensation reaction on dicarboxylic acid + diol
+            vinyl: performs addition reaction vinyl or acrylate groups
+            ester: performs condensation reaction on diacid + diol or  hydroxyacid
             carbonate: performs condensation reaction on carbonate + diol
+            amide: performs condensation reaction on diacid + diamine or aminoacid
+            urethane: performs condensation reaction on diisocyanate + diamine
+            NIPU: performs condensation reaction multi-functional cyclic carbonate + diamine
+            imide: performs condensation reaction on diacidanhydride and diamine
+
+            all: will perform all mechanims, may cause multiple reaction mechanisms if not specified
 
         replicate_structures: integer, optional
             number of replicate structures which will be generated
@@ -481,18 +497,23 @@ class PolyMaker:
             returnpoly = pd.concat([returnpoly, returnpoly_i])
 
         if verbose:
-            returnpoly[["smiles_polymer", "mechanism"]] = returnpoly.progress_apply(
+            apply_method = returnpoly.progress_apply
+        else:
+            apply_method = returnpoly.apply
+
+        if (returnpoly.keys() == "mechanism").any() & ((mechanism == "")):
+            returnpoly[["smiles_polymer", "mechanism"]] = apply_method(
                 lambda row: self.__polymerizemechanism_thermoplastic(
                     ast.literal_eval(row.monomers),
                     DP,
-                    mechanism,
+                    row.mechanism,
                     ast.literal_eval(row.distribution),
                     infinite_chain,
                 ),
                 axis=1,
             )
         else:
-            returnpoly[["smiles_polymer", "mechanism"]] = returnpoly.apply(
+            returnpoly[["smiles_polymer", "mechanism"]] = apply_method(
                 lambda row: self.__polymerizemechanism_thermoplastic(
                     ast.literal_eval(row.monomers),
                     DP,
@@ -502,10 +523,11 @@ class PolyMaker:
                 ),
                 axis=1,
             )
+
         returnpoly = returnpoly.sort_index().sort_values("replicate_structure")
         return returnpoly
 
-    def get_functionality(self, reactants, distribution=[]):
+    def get_functionality(self, reactants, distribution=[], verbose=0):
         """gets the functional groups from a list of reactants
 
         inputs: list of smiles
@@ -514,53 +536,19 @@ class PolyMaker:
 
         def id_functionality(r):
             mol = Chem.MolFromSmiles(r.name)
-            r.ols = len(
-                mol.GetSubstructMatches(Chem.MolFromSmarts(self.smiles_req["ols"]))
-            )
-            r.aliphatic_ols = len(
-                mol.GetSubstructMatches(
-                    Chem.MolFromSmarts(self.smiles_req["aliphatic_ols"])
-                )
-            )
-            r.acids = len(
-                mol.GetSubstructMatches(Chem.MolFromSmarts(self.smiles_req["acids"]))
-            )
-            r.prime_amines = len(
-                mol.GetSubstructMatches(
-                    Chem.MolFromSmarts(self.smiles_req["prime_amines"])
-                )
-            )
-            r.carbonates = len(
-                mol.GetSubstructMatches(
-                    Chem.MolFromSmarts(self.smiles_req["carbonates"])
-                )
-            )
-            r.acidanhydrides = len(
-                mol.GetSubstructMatches(
-                    Chem.MolFromSmarts(self.smiles_req["acidanhydrides"])
-                )
-            )
-            r.cyclic_carbonates = len(
-                mol.GetSubstructMatches(
-                    Chem.MolFromSmarts(self.smiles_req["cyclic_carbonates"])
-                )
-            )
+            for k, v in self.smiles_templates.items():
+                r[k] = len(mol.GetSubstructMatches(Chem.MolFromSmarts(v)))
             return r
 
         df_func = pd.DataFrame(
             data=0,
             index=reactants,
-            columns=[
-                "ols",
-                "acids",
-                "prime_amines",
-                "carbonates",
-                "aliphatic_ols",
-                "acidanhydrides",
-                "cyclic_carbonates",
-            ],
+            columns=self.smiles_templates.keys(),
         )
-        df_func = df_func.apply(lambda r: id_functionality(r), axis=1)
+        if verbose == 0:
+            df_func = df_func.apply(lambda r: id_functionality(r), axis=1)
+        if verbose >= 1:
+            df_func = df_func.progress_apply(lambda r: id_functionality(r), axis=1)
 
         # appends distribution to dataframe
 
@@ -654,6 +642,11 @@ class PolyMaker:
             returnpoly = polydata[0]
             mechanism = polydata[1]
 
+        elif mechanism == "urethane":
+            polydata = self.__poly_urethane(reactants, DP, distribution, infinite_chain)
+            returnpoly = polydata[0]
+            mechanism = polydata[1]
+
         elif mechanism == "all":
             polylist = [
                 self.__poly_vinyl(reactants, DP, distribution, infinite_chain),
@@ -662,6 +655,7 @@ class PolyMaker:
                 self.__poly_carbonate(reactants, DP, distribution, infinite_chain),
                 self.__poly_imide(reactants, DP, distribution, infinite_chain),
                 self.__poly_NIPU(reactants, DP, distribution, infinite_chain),
+                self.__poly_urethane(reactants, DP, distribution, infinite_chain),
             ]
 
             polylist = [
@@ -675,6 +669,7 @@ class PolyMaker:
                     "ERROR:Carbonate_ReactionFailed",
                     "ERROR:Imide_ReactionFailed",
                     "ERROR:NIPU_ReactionFailed",
+                    "ERROR:Urethane_ReactionFailed",
                     "",
                 ]
             ]
@@ -909,7 +904,7 @@ class PolyMaker:
                 if (
                     len(
                         mol.GetSubstructMatches(
-                            Chem.MolFromSmarts(self.smiles_req["acidanhydrides"])
+                            Chem.MolFromSmarts(self.smiles_templates["acidanhydrides"])
                         )
                     )
                     > 0
@@ -1519,6 +1514,113 @@ class PolyMaker:
 
         return poly, "NIPU"
 
+    def __poly_urethane(self, reactants, DP=2, distribution=[], infinite_chain=False):
+        """performs condenstation reaction on diisocyanates and  diols"""
+        # function
+
+        try:
+            # initial
+            rxn_dic = self.reactions["urethane"]
+            df_func = self.get_functionality(reactants, distribution=distribution)
+
+            # select initial monomer as polymer chain
+            df_poly = df_func.sample(1)
+            df_func.loc["smiles_polymer"] = df_poly.sample(1).values[0]
+            poly = df_poly.index[0]
+            molpoly = Chem.MolFromSmiles(poly)
+
+            DP_count = 1
+            DP_actual = 1
+
+            while DP_count < DP:
+
+                # select rxn rule and reactant
+                if (df_func.loc["smiles_polymer", "isocyanates"] >= 1) & (
+                    df_func.loc["smiles_polymer", "ols"] >= 1
+                ):
+                    msk = ((df_func.ols >= 1) | (df_func.isocyanates >= 1)) & (
+                        df_func.index != "smiles_polymer"
+                    )
+                    df_func_select = df_func.loc[msk]
+                    a = df_func_select.sample(
+                        1, weights=df_func.distribution, replace=True
+                    ).index.values[0]
+                    if df_func.loc[a].isocyanates >= 1:
+                        rxn_selector = "diols_isocyanate"
+                    if df_func.loc[a].ols >= 1:
+                        rxn_selector = "diisocyanates_ol"
+                elif df_func.loc["smiles_polymer", "isocyanates"] >= 2:
+                    msk = (df_func.ols >= 1) & (df_func.index != "smiles_polymer")
+                    df_func_select = df_func.loc[msk]
+                    a = df_func_select.sample(
+                        1, weights=df_func.distribution, replace=True
+                    ).index.values[0]
+                    rxn_selector = "diisocyanates_ol"
+                elif df_func.loc["smiles_polymer", "ols"] >= 2:
+                    msk = (df_func.isocyanates >= 1) & (
+                        df_func.index != "smiles_polymer"
+                    )
+                    df_func_select = df_func.loc[msk]
+                    a = df_func_select.sample(
+                        1, weights=df_func.distribution, replace=True
+                    ).index.values[0]
+                    rxn_selector = "diols_isocyanate"
+                else:
+                    assert False
+
+                rxn = Chem.AllChem.ReactionFromSmarts(rxn_dic[rxn_selector])
+
+                # update df_func table
+                df_func.loc["smiles_polymer"] = (
+                    df_func.loc["smiles_polymer"] + df_func.loc[a]
+                )  # adding polymer and a
+                for column_name in ["isocyanates", "ols"]:
+                    df_func.loc["smiles_polymer", column_name] += -1
+                assert (
+                    df_func.loc["smiles_polymer"][
+                        df_func.loc["smiles_polymer"] > -1
+                    ].shape
+                    == df_func.loc["smiles_polymer"].shape
+                )
+
+                # React and select product
+                mola = Chem.MolFromSmiles(a)
+                prod = rxn.RunReactants((molpoly, mola))
+                prodlist = [Chem.MolToSmiles(x[0]) for x in prod]
+                prodlist = self._returnvalid(prodlist)
+                poly = random.choice(prodlist)
+                molpoly = Chem.MolFromSmiles(poly)
+
+                # manage loop and ring close
+                if (infinite_chain) & (DP_count == DP - 1):
+                    # logic for closing ring
+                    if (df_func.loc["smiles_polymer", "isocyanates"] > 0) & (
+                        df_func.loc["smiles_polymer", "ols"]
+                    ) > 0:
+                        # case for when can ring close
+                        DP_count += 1
+                        DP_actual += 1
+                    else:
+                        # case for when has same terminal ends so can't ring close
+                        DP_count = DP_count
+                        DP_actual += 1
+                else:
+                    DP_count += 1
+                    DP_actual += 1
+
+            if infinite_chain:  # closes ring
+
+                rxn = Chem.AllChem.ReactionFromSmarts(rxn_dic["infinite_chain"])
+                prod = rxn.RunReactants((molpoly,))
+                prodlist = [Chem.MolToSmiles(x[0]) for x in prod]
+                prodlist = self._returnvalid(prodlist)
+                poly = random.choice(prodlist)
+                molpoly = Chem.MolFromSmiles(poly)
+        except:
+            poly = "ERROR:Urethane_ReactionFailed"
+
+        return poly, "urethane"
+
     def __poly_upe(self, reactants, crosslinker, distribution, DP):
         """generates 2 ringed thermoset
         reactants: list of smiles
